#include<Complex.h>
#include<stdlib.h>

#include"LibNic.h"
#include"Integrate.h"

#include"Symmetry.h"
#include"Green.h"

/*****************************************************************************/

EComponent::EComponent(int x, int y, int z)

{ N=0;
  if (x) {X=1; N++;} else {X=0;}
  if (y) {Y=1; N++;} else {Y=0;}
  if (z) {Z=1; N++;} else {Z=0;}
}

EComponent::EComponent(const EComponent& ec)

{ N=ec.N;
  X=ec.X;
  Y=ec.Y;
  Z=ec.Z;
}

/*****************************************************************************/

MeshGeometry::MeshGeometry(double dx, double dy, double dz)
  : Triplet<double>(dx,dy,dz)
{}; 

MeshGeometry::MeshGeometry(const MeshGeometry& mg)
  : Triplet<double>(mg)
{
}; 

double MeshGeometry::Volume() const

{  return x*y*z;
}

/*****************************************************************************/

GreenTensor::GreenTensor(Complex kref,Complex epsref,const MeshGeometry& mg, 
                         const Symmetry& s, const EComponent& ec)

{ Kref=kref;
  EpsilonRef=epsref;
  MGeom=mg;
  Symm=s;
  EComp=ec;
}

Tensor GreenTensor::GreenE0(vect3d,vect3d)

{ FatalError("Use of the uninstantiated function GreenTensor::GreenE0");
  return Tensor();
}

Tensor GreenTensor::GreenM0(vect3d,vect3d)

{ FatalError("Use of the uninstantiated function GreenTensor::GreenM0");
  return Tensor();
}

Tensor GreenTensor::GreenFFE0(vect3d,vect3d)

{ FatalError("Use of the uninstantiated function GreenTensor::GreenFFE0");
  return Tensor();
}

Tensor GreenTensor::L()

{ FatalError("Use of the uninstantiated function GreenTensor::L");
  return Tensor();
}

Complex GreenTensor::MScalarFctMGeom(const MeshGeometry&)

{ FatalError("Use of the uninstantiated function "
             "GreenTensor::MScalarFctMGeom");
  return Complex();
}

Tensor GreenTensor::GreenE(vect3d r1,vect3d r2)

{ Tensor G=GreenE0(r1,r2);

  if (!((Symm.sx)||(Symm.sy)||(Symm.sz))) return G;

  if (Symm.sx) 
    {
    G+=GreenE0(r1,vect3d(-r2.x, r2.y, r2.z))*Symm.SX();
    }
  if (Symm.sy)
    {
    G+=GreenE0(r1,vect3d( r2.x,-r2.y, r2.z))*Symm.SY();
    }
  if (Symm.sz) 
    {
    G+=GreenE0(r1,vect3d( r2.x, r2.y,-r2.z))*Symm.SZ();
    }
  if ((Symm.sx)&&(Symm.sy))  
    {
    G+=GreenE0(r1,vect3d(-r2.x,-r2.y, r2.z))*Symm.SX()*Symm.SY();
    }
  if ((Symm.sx)&&(Symm.sz)) 
    {
    G+=GreenE0(r1,vect3d(-r2.x, r2.y,-r2.z))*Symm.SX()*Symm.SZ();
    }
  if ((Symm.sy)&&(Symm.sz)) 
    {
    G+=GreenE0(r1,vect3d( r2.x,-r2.y,-r2.z))*Symm.SY()*Symm.SZ();
    }
  if ((Symm.sx)&&(Symm.sy)&&(Symm.sz)) 
    {
    G+=GreenE0(r1,vect3d(-r2.x,-r2.y,-r2.z))*Symm.SX()*Symm.SY()*Symm.SZ();
    }
  return G;
}

Tensor GreenTensor::GreenM(vect3d r1,vect3d r2)

{ Tensor G=GreenM0(r1,r2);

  if (!((Symm.sx)||(Symm.sy)||(Symm.sz))) return G;

  if (Symm.sx) 
    {
    G+=GreenM0(r1,vect3d(-r2.x, r2.y, r2.z))*Symm.SX();
    }
  if (Symm.sy)
    {
    G+=GreenM0(r1,vect3d( r2.x,-r2.y, r2.z))*Symm.SY();
    }
  if (Symm.sz) 
    {
    G+=GreenM0(r1,vect3d( r2.x, r2.y,-r2.z))*Symm.SZ();
    }
  if ((Symm.sx)&&(Symm.sy))  
    {
    G+=GreenM0(r1,vect3d(-r2.x,-r2.y, r2.z))*Symm.SX()*Symm.SY();
    }
  if ((Symm.sx)&&(Symm.sz)) 
    {
    G+=GreenM0(r1,vect3d(-r2.x, r2.y,-r2.z))*Symm.SX()*Symm.SZ();
    }
  if ((Symm.sy)&&(Symm.sz)) 
    {
    G+=GreenM0(r1,vect3d( r2.x,-r2.y,-r2.z))*Symm.SY()*Symm.SZ();
    }
  if ((Symm.sx)&&(Symm.sy)&&(Symm.sz)) 
    {
    G+=GreenM0(r1,vect3d(-r2.x,-r2.y,-r2.z))*Symm.SX()*Symm.SY()*Symm.SZ();
    }
  return G;
}
Tensor GreenTensor::GreenFFE(vect3d n1,vect3d r2)

{ Tensor G=GreenFFE0(n1,r2);

  if (!((Symm.sx)||(Symm.sy)||(Symm.sz))) return G;

  if (Symm.sx) 
    {
    G+=GreenFFE0(n1,vect3d(-r2.x, r2.y, r2.z))*Symm.SX();
    }
  if (Symm.sy)
    {
    G+=GreenFFE0(n1,vect3d( r2.x,-r2.y, r2.z))*Symm.SY();
    }
  if (Symm.sz) 
    {
    G+=GreenFFE0(n1,vect3d( r2.x, r2.y,-r2.z))*Symm.SZ();
    }
  if ((Symm.sx)&&(Symm.sy))  
    {
    G+=GreenFFE0(n1,vect3d(-r2.x,-r2.y, r2.z))*Symm.SX()*Symm.SY();
    }
  if ((Symm.sx)&&(Symm.sz)) 
    {
    G+=GreenFFE0(n1,vect3d(-r2.x, r2.y,-r2.z))*Symm.SX()*Symm.SZ();
    }
  if ((Symm.sy)&&(Symm.sz)) 
    {
    G+=GreenFFE0(n1,vect3d( r2.x,-r2.y,-r2.z))*Symm.SY()*Symm.SZ();
    }
  if ((Symm.sx)&&(Symm.sy)&&(Symm.sz)) 
    {
    G+=GreenFFE0(n1,vect3d(-r2.x,-r2.y,-r2.z))*Symm.SX()*Symm.SY()*Symm.SZ();
    }
  return G;
}

/*****************************************************************************/

Green3dBasicClass::Green3dBasicClass(Complex kref,Complex epsref, 
                                     const MeshGeometry& mg,
                                     const Symmetry& s,const EComponent& ec) :
  GreenTensor(kref,epsref,mg,s,ec)

{ if (EComp.N!=3) FatalError("Number of field component uncompatible with "
                             "the 3D green function");
}

Tensor Green3dBasicClass::L()

{ Tensor T=Tensor(eyes)*(1.0/3.0);
  return T;
}

Tensor Green3dBasicClass::GreenFFE0(vect3d n1,vect3d Rsource)
// New definition

{ double rn=Rsource*n1;

  Tensor T=1./4./PI*(Tensor(eyes)-prod(n1,n1)) * exp(-Complex(0,1)*Kref*rn);
  return T;
}

/*****************************************************************************/

Green3d::Green3d(Complex kref,Complex epsref, const MeshGeometry& mg, 
                 const Symmetry& s,
                 const EComponent& ec) :
  Green3dBasicClass(kref,epsref,mg,s,ec)

{}

Green3d::Green3d(Complex kref,Complex epsref, double MeshSize) 
  : Green3dBasicClass(kref,epsref,MeshGeometry(MeshSize,MeshSize,MeshSize),
                      Symmetry(0,0,0),EComponent(1,1,1))

{}

Tensor Green3d::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ Tensor T=Tensor(zeros);
  Tensor I=Tensor(eyes);
  Complex i=Complex(0,1);

  double rabs=Abs(r1-Rsource);

  if(rabs>NullDistance)
    {
    vect3d R=r1-Rsource;

    Complex F   =exp(i*Kref*rabs)/(4*PI*rabs);
    Complex DF  =(i*Kref - 1.0/rabs) * F;
    Complex DDF =pow(rabs,-2) * F + (i*Kref - 1.0/rabs) * DF;

    T=(F+DF/Kref/Kref/rabs)*I + 
      (-DF/Kref/Kref/pow(rabs,3) + DDF/Kref/Kref/pow(rabs,2) )* prod(R,R); 
    }
  else FatalError("Evaluation of Green's tensor on its singularity");
    
  return T;
}

Complex Green3d::MScalarFctMGeom(const MeshGeometry& mgeom)
// New definition

{ double RSphere=mgeom.REq3D();
  Complex i=Complex(0,1);

  return 2.0/(3.0*pow(Kref,2))* (-1+(1-i*Kref*RSphere)*exp(i*Kref*RSphere));

}

Tensor Green3d::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double rabs=Abs(r1-Rsource);
  vect3d R=r1-Rsource;
  Tensor T=Tensor(zeros);
  Complex i=Complex(0,1);

  T=Tensor(vect3dc( 0,   -R.z,  R.y),
           vect3dc( R.z,  0,   -R.x),
           vect3dc(-R.y,  R.x,  0  ));

  Complex F   =exp(i*Kref*rabs)/(4*PI*rabs);
  Complex DF  =(i*Kref - 1.0/rabs) * F;

//  return i/Kref*sqrt(EpsilonRef*EPSILONo/MUo)*
//         exp(i*Kref*rabs)*(1-i*Kref*rabs)/(4*PI*rabs*rabs*rabs)*T;

  if (rabs==0) return 0*T;

  Tensor TT=-i/Kref*sqrt(EpsilonRef*EPSILONo/MUo)*DF/rabs*T;

  return TT;
}


/*****************************************************************************/

FGreen3d::FGreen3d(Complex kref,Complex epsref,double kf, 
                   const MeshGeometry& mg,
                   const Symmetry& s, const EComponent& ec) :
 Green3dBasicClass(kref,epsref,mg,s,ec)

{ KF=kf;

  if(Abs(arg(kref))>0.01)
    FatalError("Complex KRef not implemnted for Green3dF");
}

FGreen3d::FGreen3d(Complex kref,Complex epsref,double MeshSize) :
 Green3dBasicClass(kref,epsref,MeshGeometry(MeshSize,MeshSize,MeshSize),
                      Symmetry(0,0,0),EComponent(1,1,1))

{ KF=PI/MeshSize;
}

Tensor FGreen3d::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ Tensor T=Tensor(zeros);
  Tensor I=Tensor(eyes);
  Complex i=Complex(0,1);

  double rabs=Abs(r1-Rsource);
  double HImp=0;
 
  if(rabs>NullDistance)
    {
    vect3d R=vect3d(r1-Rsource);
    Complex k2r2=pow(Kref*rabs,2);

    double CIP= CosIntegral((KF + real(Kref))*rabs);
    double SIP= SinIntegral((KF + real(Kref))*rabs);
    double CIM= CosIntegral((KF - real(Kref))*rabs);
    double SIM= SinIntegral((KF - real(Kref))*rabs);

    Complex F, DF, DDF;

      F= 
        -(-exp(Complex(0,1)*Kref*rabs)/(4*PI*rabs) + 
        ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
        (4*pow(PI,2)*rabs));

      DF=
        -(exp(Complex(0,1)*Kref*rabs)/(4*PI*pow(rabs,2)) - 
        Complex(0,1.0/4.0)*Kref*exp(Complex(0,1)*Kref*rabs)/(PI*rabs) - 
        ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
        (4*pow(PI,2)*pow(rabs,2)) + 
        (-(CIM*Kref*cos(Kref*rabs)) + CIP*Kref*cos(Kref*rabs) - 
        2*sin(KF*rabs)/rabs - Kref*PI*sin(Kref*rabs) + 
        Kref*SIM*sin(Kref*rabs) + Kref*SIP*sin(Kref*rabs))/(4*pow(PI,2)*rabs));

      DDF=
        -(-exp(Complex(0,1)*Kref*rabs)/(2*PI*pow(rabs,3)) + 
        Complex(0,1.0/2.0)*Kref*exp(Complex(0,1)*Kref*rabs)/(PI*pow(rabs,2)) + 
        pow(Kref,2)*exp(Complex(0,1)*Kref*rabs)/(4*PI*rabs) + 
        ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
        (2*pow(PI,2)*pow(rabs,3)) + 
        (-2*KF*cos(KF*rabs)/rabs - pow(Kref,2)*PI*cos(Kref*rabs) + 
        pow(Kref,2)*SIM*cos(Kref*rabs) + pow(Kref,2)*SIP*cos(Kref*rabs) + 
        2*sin(KF*rabs)/pow(rabs,2) + CIM*pow(Kref,2)*sin(Kref*rabs) - 
        CIP*pow(Kref,2)*sin(Kref*rabs))/(4*pow(PI,2)*rabs) + 
        (CIM*Kref*cos(Kref*rabs) - CIP*Kref*cos(Kref*rabs) + 
        2*sin(KF*rabs)/rabs + Kref*PI*sin(Kref*rabs) - 
        Kref*SIM*sin(Kref*rabs) - Kref*SIP*sin(Kref*rabs))/
        (2*pow(PI,2)*pow(rabs,2)));
     
      HImp=4.0*PI*(sin(KF*rabs)/rabs-KF*cos(KF*rabs))/rabs/rabs;

    T=(F+DF/Kref/Kref/rabs)*I + 
      (-DF/Kref/Kref/pow(rabs,3) + DDF/Kref/Kref/pow(rabs,2) )* (prod(R,R))//;
       +1.0/3.0/pow(Kref,2)*HImp/pow(2*PI,3)*I ;
    } 

  else 
    {
    //FatalError("Evaluation of Green's tensor on its singularity");
    return MScalar()/MGeom.Volume()*Tensor(eyes);
    }

  return T;
}

Tensor FGreen3d::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double rabs=Abs(r1-Rsource);
  vect3d R=r1-Rsource;
  Tensor T=Tensor(zeros);
  Complex i=Complex(0,1);
  Complex k2r2=pow(Kref*rabs,2);

  T=Tensor(vect3dc( 0,   -R.z,  R.y),
           vect3dc( R.z,  0,   -R.x),
           vect3dc(-R.y,  R.x,  0  ));

    double CIP= CosIntegral((KF + real(Kref))*rabs);
    double SIP= SinIntegral((KF + real(Kref))*rabs);
    double CIM= CosIntegral((KF - real(Kref))*rabs);
    double SIM= SinIntegral((KF - real(Kref))*rabs);

    Complex DF=
      -(exp(Complex(0,1)*Kref*rabs)/(4*PI*pow(rabs,2)) - 
      Complex(0,1.0/4.0)*Kref*exp(Complex(0,1)*Kref*rabs)/(PI*rabs) - 
      ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
      (4*pow(PI,2)*pow(rabs,2)) + 
      (-(CIM*Kref*cos(Kref*rabs)) + CIP*Kref*cos(Kref*rabs) - 
      2*sin(KF*rabs)/rabs - Kref*PI*sin(Kref*rabs) + 
      Kref*SIM*sin(Kref*rabs) + Kref*SIP*sin(Kref*rabs))/(4*pow(PI,2)*rabs));

  if (rabs==0) return 0*T;

  return -i/Kref*sqrt(EpsilonRef*EPSILONo/MUo)*DF/rabs*T;
}

Complex FGreen3d::MScalar()
// New definition

{ Complex i=Complex(0,1);

  Complex l1=log(KF - Kref);
  Complex l2=log(KF + Kref);

  Complex P1=
  -((pow(KF,3) - 6*KF*pow(Kref,2) - Complex(0,3)*pow(Kref,3)*PI - 
    3*pow(Kref,3)*log(KF - Kref) + 3*pow(Kref,3)*log(KF + Kref))/
  (18*pow(Kref,2)*pow(PI,2)));
 
  double HImp=4.0*PI*pow(KF,3)/3.0;

  Complex T=P1;
  T+= 1.0/3.0/pow(Kref,2)*HImp/pow(2*PI,3);

  return T*MGeom.Volume();
}



/*****************************************************************************/
/*****************************************************************************/

FGreen3dL::FGreen3dL(Complex kref,Complex epsref,double kf, 
                     const MeshGeometry& mg,
                     const Symmetry& s, const EComponent& ec) :
 Green3dBasicClass(kref,epsref,mg,s,ec)

{ KF=kf;

  if(Abs(arg(kref))>0.01)
    FatalError("Complex KRef not implemnted for Green3dF");
}

Tensor FGreen3dL::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ Tensor T=Tensor(zeros);
  Tensor I=Tensor(eyes);
  Complex i=Complex(0,1);

  double rabs=Abs(r1-Rsource);

  if(rabs>NullDistance)
    {
    vect3d R=vect3d(r1-Rsource);
    Complex k2r2=pow(Kref*rabs,2);

    double CIP= CosIntegral((KF + real(Kref))*rabs);
    double SIP= SinIntegral((KF + real(Kref))*rabs);
    double CIM= CosIntegral((KF - real(Kref))*rabs);
    double SIM= SinIntegral((KF - real(Kref))*rabs);

    Complex F, DF, DDF;

      F= 
        -(-exp(Complex(0,1)*Kref*rabs)/(4*PI*rabs) + 
        ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
        (4*pow(PI,2)*rabs));

      DF=
        -(exp(Complex(0,1)*Kref*rabs)/(4*PI*pow(rabs,2)) - 
        Complex(0,1.0/4.0)*Kref*exp(Complex(0,1)*Kref*rabs)/(PI*rabs) - 
        ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
        (4*pow(PI,2)*pow(rabs,2)) + 
        (-(CIM*Kref*cos(Kref*rabs)) + CIP*Kref*cos(Kref*rabs) - 
        2*sin(KF*rabs)/rabs - Kref*PI*sin(Kref*rabs) + 
        Kref*SIM*sin(Kref*rabs) + Kref*SIP*sin(Kref*rabs))/(4*pow(PI,2)*rabs));

      DDF=
        -(-exp(Complex(0,1)*Kref*rabs)/(2*PI*pow(rabs,3)) + 
        Complex(0,1.0/2.0)*Kref*exp(Complex(0,1)*Kref*rabs)/(PI*pow(rabs,2)) + 
        pow(Kref,2)*exp(Complex(0,1)*Kref*rabs)/(4*PI*rabs) + 
        ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
        (2*pow(PI,2)*pow(rabs,3)) + 
        (-2*KF*cos(KF*rabs)/rabs - pow(Kref,2)*PI*cos(Kref*rabs) + 
        pow(Kref,2)*SIM*cos(Kref*rabs) + pow(Kref,2)*SIP*cos(Kref*rabs) + 
        2*sin(KF*rabs)/pow(rabs,2) + CIM*pow(Kref,2)*sin(Kref*rabs) - 
        CIP*pow(Kref,2)*sin(Kref*rabs))/(4*pow(PI,2)*rabs) + 
        (CIM*Kref*cos(Kref*rabs) - CIP*Kref*cos(Kref*rabs) + 
        2*sin(KF*rabs)/rabs + Kref*PI*sin(Kref*rabs) - 
        Kref*SIM*sin(Kref*rabs) - Kref*SIP*sin(Kref*rabs))/
        (2*pow(PI,2)*pow(rabs,2)));
     
    T=(F+DF/Kref/Kref/rabs)*I + 
      (-DF/Kref/Kref/pow(rabs,3) + DDF/Kref/Kref/pow(rabs,2) )* (prod(R,R));
    } 

  else 
    {
    //FatalError("Evaluation of Green's tensor on its singularity");
    T=MScalar()/MGeom.Volume()*Tensor(eyes);
    }

  return T;
}

Tensor FGreen3dL::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double rabs=Abs(r1-Rsource);
  vect3d R=r1-Rsource;
  Tensor T=Tensor(zeros);
  Complex i=Complex(0,1);
  Complex k2r2=pow(Kref*rabs,2);

  T=Tensor(vect3dc( 0,   -R.z,  R.y),
           vect3dc( R.z,  0,   -R.x),
           vect3dc(-R.y,  R.x,  0  ));

    double CIP= CosIntegral((KF + real(Kref))*rabs);
    double SIP= SinIntegral((KF + real(Kref))*rabs);
    double CIM= CosIntegral((KF - real(Kref))*rabs);
    double SIM= SinIntegral((KF - real(Kref))*rabs);

    Complex DF=
      -(exp(Complex(0,1)*Kref*rabs)/(4*PI*pow(rabs,2)) - 
      Complex(0,1.0/4.0)*Kref*exp(Complex(0,1)*Kref*rabs)/(PI*rabs) - 
      ((PI - SIM - SIP)*cos(Kref*rabs) + (-CIM + CIP)*sin(Kref*rabs))/
      (4*pow(PI,2)*pow(rabs,2)) + 
      (-(CIM*Kref*cos(Kref*rabs)) + CIP*Kref*cos(Kref*rabs) - 
      2*sin(KF*rabs)/rabs - Kref*PI*sin(Kref*rabs) + 
      Kref*SIM*sin(Kref*rabs) + Kref*SIP*sin(Kref*rabs))/(4*pow(PI,2)*rabs));

  if (rabs==0) return 0*T;

  return -i/Kref*sqrt(EpsilonRef*EPSILONo/MUo)*DF/rabs*T;
}

Complex FGreen3dL::MScalar()
// New definition

{ Complex i=Complex(0,1);

  Complex l1=log(KF - Kref);
  Complex l2=log(KF + Kref);

  Complex P1=
  -((pow(KF,3) - 6*KF*pow(Kref,2) - Complex(0,3)*pow(Kref,3)*PI - 
    3*pow(Kref,3)*log(KF - Kref) + 3*pow(Kref,3)*log(KF + Kref))/
  (18*pow(Kref,2)*pow(PI,2)));

  return P1*MGeom.Volume();
}

Tensor FGreen3dL::L()

{ return Tensor(zeros);
}


/*****************************************************************************/



Complex Hankel1Inf(int order, Complex k, double rn)

{ Complex I=Complex(0,1);

  return sqrt(2/PI/k)*exp(-I*k*rn)*exp(-I*(2.0*order-1.0)/4.0*PI);
}

/*****************************************************************************/
/*****************************************************************************/

Green2dTMBasicClass::Green2dTMBasicClass(Complex kref,Complex epsref,
                                     const MeshGeometry& mg, 
                                     const Symmetry& s,const EComponent& ec) :

 GreenTensor(kref,epsref,mg,s,ec)

{ if (EComp.N!=1) FatalError("Number of field component uncompatible with "
                             "the 3D green function");

  if (!(EComp.Z)) FatalError("Uncompatible Component for te Case");
}

Tensor Green2dTMBasicClass::L()

{ return Tensor(zeros);
}

Tensor Green2dTMBasicClass::GreenFFE0(vect3d n1,vect3d Rsource)

{ double nr=MakeVect2D(n1)*MakeVect2D(Rsource);
  Tensor G=Tensor(zeros);

  G.z.z=Complex(0,1.0/4.0)* Hankel1Inf(0,Kref,nr);
  return G;  	
}

/*****************************************************************************/

Green2dTM::Green2dTM(Complex kref,Complex epsref, const MeshGeometry& mg, const Symmetry& s,
                       const EComponent& ec) :
 Green2dTMBasicClass(kref,epsref,mg,s,ec)

{ 
}

Green2dTM::Green2dTM(Complex kref,Complex epsref, double MeshSize) 
 :
 Green2dTMBasicClass(kref,epsref,MeshGeometry(MeshSize,MeshSize,1),Symmetry(0,0,0),EComponent(0,0,1))

{ 
}


Tensor Green2dTM::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));
  Tensor G=Tensor(zeros);

  if (R>NullDistance) G.z.z=Complex(0,1.0/4.0)* Hankel1(0,Kref*R);
  else 
    {
    FatalError("Evaluation of Green's tensor on its singularity");
    }

  return G;  	
}

Complex Green2dTM::MScalarFctMGeom(const MeshGeometry& mgeom)
// New definition

{   double Req=mgeom.REq2D();
    Complex I=Complex(0,1);
    
    return (I*PI/2/Kref*Req*Hankel1(1,Kref*Req)-1/Kref/Kref);
}

Tensor Green2dTM::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));

  if (R>NullDistance) 
    {
    vect2d r=MakeVect2D(r1-Rsource)/R;
    Tensor GM=Tensor(zeros);

    Complex C=1.0/4.0*sqrt(EpsilonRef*EPSILONo/MUo)*Hankel1(1,Kref*R);
    
    GM.x.z=-C*r.y;
    GM.y.z= C*r.x;
    return GM;
    }
  else return Tensor(zeros);  	
}

/*****************************************************************************/

FGreen2dTM::FGreen2dTM(Complex kref,Complex epsref, double kf,
                       const MeshGeometry& mg,
                       const Symmetry& s,const EComponent& ec) :
 Green2dTMBasicClass(kref,epsref,mg,s,ec)

{ KF=kf;
}

FGreen2dTM::FGreen2dTM(Complex kref,Complex epsref, double MeshSize) 
 :
 Green2dTMBasicClass(kref,epsref,MeshGeometry(MeshSize,MeshSize,1),Symmetry(0,0,0),EComponent(0,0,1))

{ KF=PI/MeshSize;
}

Tensor FGreen2dTM::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));
  Tensor G=Tensor(zeros);
  FiltHankel1 FH=FiltHankel1(Kref,KF);

  if(R>NullDistance) G.z.z=Complex(0,1.0/4.0)* FH.FHankelOrder0(R);
  else  
    {
    //FatalError("Evaluation of Green's tensor on its singularity");
    G=MScalar()/MGeom.Volume()*Tensor(eyes);
    }

  return G;  	
}

Complex FGreen2dTM::MScalar()
// New definition

{ FiltHankel1 FH=FiltHankel1(Kref,KF);

  return Complex(0,1.0/4.0)* FH.FHankelOrder0(0)*MGeom.Volume();  	
}

Tensor FGreen2dTM::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));
  FiltHankel1 FH=FiltHankel1(Kref,KF);

  vect2d r=MakeVect2D(r1-Rsource)/R;
  Tensor GM=Tensor(zeros);

  Complex C=1.0/4.0*sqrt(EpsilonRef*EPSILONo/MUo)*FH.FHankelOrder1(R);
    
  GM.x.z=-C*r.y;
  GM.y.z= C*r.x;
  return GM;
}

/*****************************************************************************/

Green2dTEBasicClass::Green2dTEBasicClass(Complex kref,Complex epsref,
                      const MeshGeometry& mg,
                      const Symmetry& s, const EComponent& ec) :
 GreenTensor(kref,epsref,mg,s,ec)

{ if (EComp.N!=2) FatalError("Number of field component uncompatible with "
                             "the 2D TE green function");

  if (!(EComp.X)) FatalError("Uncompatible Component for TE Case");
  if (!(EComp.Y)) FatalError("Uncompatible Component for TE Case");
}

Tensor Green2dTEBasicClass::L()

{ Tensor G=Tensor(zeros);

  G.x.x=0.5;
  G.y.y=0.5;
  return G;
}

Tensor Green2dTEBasicClass::GreenFFE0(vect3d n1,vect3d Rsource)
// New definition

{ double nr=MakeVect2D(n1)*MakeVect2D(Rsource);
  Tensor G=Tensor(zeros);

  Complex H0=Hankel1Inf(0,Kref,nr);
  Complex H2=Hankel1Inf(2,Kref,nr);
  double Phi=atan2(n1.y, n1.x);
    
  G.x.x=Complex(0,0.25)*0.5*(H0 + H2*cos(2*Phi));
  G.y.y=Complex(0,0.25)*0.5*(H0 - H2*cos(2*Phi));
  G.x.y=Complex(0,0.25)*0.5* H2*sin(2*Phi);
  G.y.x=Complex(0,0.25)*0.5* H2*sin(2*Phi);

  return G;  	
}

/*****************************************************************************/

Green2dTE::Green2dTE(Complex kref,Complex epsref, const MeshGeometry& mg ,
                     const Symmetry& s,
                     const EComponent& ec) :
 Green2dTEBasicClass(kref,epsref,mg,s,ec)

{}

Green2dTE::Green2dTE(Complex kref,Complex epsref, double MeshSize) :
 Green2dTEBasicClass(kref,epsref,MeshGeometry(MeshSize,MeshSize,1),Symmetry(0,0,0),EComponent(1,1,0))

{}

Tensor Green2dTE::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ vect2d vr=MakeVect2D(r1)-MakeVect2D(Rsource);
  double R=Abs(vr);
  Tensor G=Tensor(zeros);

  if (R>NullDistance)
    {
    Complex H0=Hankel1(0,Kref*R);
    Complex H2=Hankel1(2,Kref*R);
    double Phi=atan2(vr.y, vr.x);
    
    G.x.x=Complex(0,0.25)*0.5*(H0 + H2*cos(2*Phi));
    G.y.y=Complex(0,0.25)*0.5*(H0 - H2*cos(2*Phi));
    G.x.y=Complex(0,0.25)*0.5* H2*sin(2*Phi);
    G.y.x=Complex(0,0.25)*0.5* H2*sin(2*Phi);
    } 
  else 
    {
    FatalError("Evaluation of Green's tensor on its singularity");
    }

  return G;  	
}

Complex Green2dTE::MScalarFctMGeom(const MeshGeometry& mgeom)
// New definition

{ double Req=mgeom.REq2D();
  Complex I=Complex(0,1);

  return (I*PI/2/Kref*Req*Hankel1(1,Kref*Req)-1/Kref/Kref)/2.0;
}

Tensor Green2dTE::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));

  if (R>NullDistance) 
    {
    vect2d r=MakeVect2D(r1-Rsource)/R;
    Tensor GM=Tensor(zeros);

    Complex C=1.0/4.0*sqrt(EpsilonRef*EPSILONo/MUo)*Hankel1(1,Kref*R);
    
    GM.x.z=-C*r.y;   // !!!!!!!!!!!!!!
    GM.y.z= C*r.x;
    return GM;
    }
  else return Tensor(zeros);  	
}

/*****************************************************************************/

FGreen2dTE::FGreen2dTE(Complex kref,Complex epsref, double kf, 
                       const MeshGeometry& mg, 
                       const Symmetry& s,const EComponent& ec) :
 Green2dTEBasicClass(kref,epsref,mg,s,ec)

{ KF=kf;
}

FGreen2dTE::FGreen2dTE(Complex kref,Complex epsref, double MeshSize) :
 Green2dTEBasicClass(kref,epsref,MeshGeometry(MeshSize,MeshSize,1),Symmetry(0,0,0),EComponent(1,1,0))

{ KF=PI/MeshSize;
}


Tensor FGreen2dTE::GreenE0(vect3d r1,vect3d Rsource)
// New definition

{ vect2d vr=MakeVect2D(r1)-MakeVect2D(Rsource);
  double R=Abs(vr);
  Tensor G=Tensor(zeros);
  FiltHankel1 FH=FiltHankel1(Kref,KF);

  Complex H0=FH.FHankelOrder0(R);
  Complex H2=FH.FHankelOrder2(R);
  double Phi=atan2(vr.y, vr.x);
    
  G.x.x=Complex(0,0.25)*0.5*(H0 + H2*cos(2*Phi));
  G.y.y=Complex(0,0.25)*0.5*(H0 - H2*cos(2*Phi));
  G.x.y=Complex(0,0.25)*0.5* H2*sin(2*Phi);
  G.y.x=Complex(0,0.25)*0.5* H2*sin(2*Phi);
   
  if(!R>NullDistance)
    { 
    //FatalError("Evaluation of Green's tensor on its singularity");
    G=MScalar()/MGeom.Volume()*Tensor(eyes);
    }
  return G;  	
}

Complex FGreen2dTE::MScalar()
// New definition

{ Tensor G=Tensor(zeros);
  FiltHankel1 FH=FiltHankel1(Kref,KF);

  Complex H0=FH.FHankelOrder0(0);
    
  return Complex(0,0.25)*0.5* H0 * MGeom.Volume();;
}

Tensor FGreen2dTE::GreenM0(vect3d r1,vect3d Rsource)
// New definition

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));
  FiltHankel1 FH=FiltHankel1(Kref,KF);

  vect2d r=MakeVect2D(r1-Rsource)/R;
  Tensor GM=Tensor(zeros);

  Complex C=1.0/4.0*sqrt(EpsilonRef*EPSILONo/MUo)*FH.FHankelOrder1(R);
    
  GM.x.z=-C*r.y;
  GM.y.z= C*r.x;
  return GM;
}

/*****************************************************************************/

FiltHankel1::FiltHankel1(Complex k0,double kF)

{ K0=k0;
  KF=kF;

  if(KF<real(K0)) FatalError("(KF<K0)\n");
  T0=real(K0);
}

Complex FiltHankel1::FZ0(double u)

{ double t=T0*tan(u);
  Complex z1=Complex(KF, t);
  Complex z2=Complex(KF,-t);

  return 1/PI* (-z1/(z1*z1-K0*K0)*Hankel1(0,z1*R) 
                +z2/(z2*z2-K0*K0)*Hankel2(0,z2*R)) * (T0+t*t/T0);
}

Complex FiltHankel1::FZ1(double u)

{ double t=T0*tan(u);
  Complex z1=Complex(KF, t);
  Complex z2=Complex(KF,-t);

  return 1/PI* (-z1*z1/K0/(z1*z1-K0*K0)*Hankel1(1,z1*R) 
                +z2*z2/K0/(z2*z2-K0*K0)*Hankel2(1,z2*R)) * (T0+t*t/T0);
}

Complex FiltHankel1::FZ2(double u)

{ double t=T0*tan(u);
  Complex z1=Complex(KF, t);
  Complex z2=Complex(KF,-t);

  return 1/PI* (-z1*z1*z1/K0/K0/(z1*z1-K0*K0)*Hankel1(2,z1*R) 
                +z2*z2*z2/K0/K0/(z2*z2-K0*K0)*Hankel2(2,z2*R)) * (T0+t*t/T0);
}

Complex FiltHankel1::FHankelOrder0(double r)

{ R=r;
  Complex C;

  if (R*100<2*PI/Abs(K0))
    {
    return 1-arg(K0*K0/(KF*KF-K0*K0))/PI 
           +Complex(0,1)/PI*log(Abs(K0*K0/(KF*KF-K0*K0)));
    }
  Integrate1dComplex i1 =Integrate1dComplex();
  i1.Datas.AccRel=AccRelIntegrFiltGreen;

  C=i1.Integrate(this,(Complex(Object::*)(double))&FiltHankel1::FZ0,0,3.14/2);
  return Hankel1(0,K0*r)+C;
}

Complex FiltHankel1::FHankelOrder1(double r)

{ R=r;
  Complex C;
  if (R*300<2*PI/Abs(K0))
    {
    return
      (Complex(0,(-1.0)/2.0)*((KF*KF) - (K0*K0)*(log((K0*K0))-Complex(0,1)*PI)
         + (K0*K0)*log((KF*KF) - (K0*K0)))/(K0*PI))*R;
    }
  Integrate1dComplex i1 =Integrate1dComplex();
  i1.Datas.AccRel=AccRelIntegrFiltGreen;

  C=i1.Integrate(this,(Complex(Object::*)(double))&FiltHankel1::FZ1,0,3.14/2);
  return Hankel1(1,K0*r)+C;
}

Complex FiltHankel1::FHankelOrder2(double r)

{ R=r;
  Complex C;
  if (R*300<2*PI/Abs(K0))
    {
    return 0;   //!!!!!!!!!!!!!!
    }
  Integrate1dComplex i1 =Integrate1dComplex();
  i1.Datas.AccRel=AccRelIntegrFiltGreen;

  C=i1.Integrate(this,(Complex(Object::*)(double))&FiltHankel1::FZ2,0,3.14/2);
  return Hankel1(2,K0*r)+C;
}

/*****************************************************************************/

Green2dBasicClass::Green2dBasicClass(Complex kref,Complex epsilonRef,
     const MeshGeometry& mg, const Symmetry& s, const EComponent& ec,
     double theta) : GreenTensor(kref,epsilonRef,mg,s,ec)

{ Theta=theta;
  Kpar=kref*sin(Theta);
  Kz  =kref*cos(Theta);
}

Complex Green2dBasicClass::Hankel1_0(double r)

{ cerr<<"Green2dBasicClass::H1";
  FatalError("call of Green2dBasicClass::Hankel1_0");
  return Complex(0,r);
}

Complex Green2dBasicClass::Hankel1_1(double r)

{ FatalError("call of Green2dBasicClass::Hankel1_1");
  return Complex(0,r);
}

Complex Green2dBasicClass::Hankel1_2(double r)

{ FatalError("call of Green2dBasicClass::Hankel1_2");
  return Complex(0,r);
}

Tensor Green2dBasicClass::GreenE0(vect3d r1,vect3d Rsource)

{ vect2d vr=MakeVect2D(r1)-MakeVect2D(Rsource);

  double R=Abs(vr);
  Tensor G=Tensor(zeros);

  Complex I(0,1);

  Complex H0=Hankel1_0(Abs(R));
  Complex H1=Hankel1_1(Abs(R));
  Complex H2=Hankel1_2(Abs(R));

  Complex Kpar2 = pow(Kpar,2);
  Complex Kz2   = pow(Kz,2);
  Complex Kref2 = pow(Kref,2);


  double Phi=atan2(vr.x, vr.y);
  
  G.y.y=(1-Kpar2/(2*Kref2)) * H0  +  Kpar2/(2*Kref2) * cos(2*Phi) * H2;
  G.y.z=G.z.y=(-I*Kpar*Kz/Kref2) * cos(Phi)  * H1;
  G.y.x=G.x.y=(Kpar2/(2*Kref2))  * sin(2*Phi)* H2;
  G.z.z=(1-Kz2/Kref2) * H0;
  G.z.x=G.x.z=(-I*Kpar*Kz/Kref2) * sin(Phi)  * H1;
  G.x.x=(1-Kpar2/(2*Kref2)) * H0  -  Kpar2/(2*Kref2) * cos(2*Phi) * H2;

  Tensor T=Complex(0,0.25)*G;
  return T;  //Complex(0,0.25)*G;  	
}

Tensor Green2dBasicClass::GreenM0(vect3d r1,vect3d Rsource)

{ double R=Abs(MakeVect2D(r1)-MakeVect2D(Rsource));

  Complex H0=Hankel1_0(Abs(R));
  Complex H1=Hankel1_1(Abs(R));

  vect2d r=MakeVect2D(r1-Rsource)/R;
  Tensor GM=Tensor(zeros);

  double cosphi=r.x;
  double sinphi=r.y;

  Complex C=sqrt(EpsilonRef*EPSILONo/MUo)/(4*Kref);
  Complex I=Complex(0,1);

  GM.x.y=-I*Kz*H0;
  GM.x.z=-Kpar*H1*sinphi;
  GM.y.z= Kpar*H1*cosphi;

  GM.y.x=-GM.x.y;
  GM.z.x=-GM.x.z;
  GM.z.y=-GM.y.z;

  Tensor T=C*GM;
  return T;
}

Tensor Green2dBasicClass::GreenFFE0(vect3d n1,vect3d Rsource)

{ double nr=MakeVect2D(n1)*MakeVect2D(Rsource);
  Tensor G=Tensor(zeros);

  Complex I(0,1);

  Complex H0=Hankel1Inf(0,Kpar,nr);
  Complex H1=Hankel1Inf(1,Kpar,nr);
  Complex H2=Hankel1Inf(2,Kpar,nr);

  Complex Kpar2 = pow(Kpar,2);
  Complex Kz2   = pow(Kz,2);
  Complex Kref2 = pow(Kref,2);


  double Phi=atan2(n1.x, n1.y);
  
  G.y.y=(1-Kpar2/(2*Kref2)) * H0  +  Kpar2/(2*Kref2) * cos(2*Phi) * H2;
  G.y.z=G.z.y=(-I*Kpar*Kz/Kref2) * cos(Phi)  * H1;
  G.y.x=G.x.y=(Kpar2/(2*Kref2))  * sin(2*Phi)* H2;
  G.z.z=(1-Kz2/Kref2) * H0;
  G.z.x=G.x.z=(-I*Kpar*Kz/Kref2) * sin(Phi)  * H1;
  G.x.x=(1-Kpar2/(2*Kref2)) * H0  -  Kpar2/(2*Kref2) * cos(2*Phi) * H2;

  Tensor T=Complex(0,0.25)*G;

  return T;  	
}

Tensor Green2dBasicClass::L()

{ Tensor l(zeros);

  l.x.x=0.5;
  l.y.y=0.5;
  return l;
}

Complex Green2d::Hankel1_0(double r)

{ return Hankel1(0,Kpar*r);
}

Complex Green2d::Hankel1_1(double r)

{ return Hankel1(1,Kpar*r);
}

Complex Green2d::Hankel1_2(double r)

{ return Hankel1(2,Kpar*r);
}


Green2d::Green2d(Complex kref,Complex epsilonRef,
      const MeshGeometry& mg, const Symmetry& s, const EComponent& ec,
      double theta)

  : Green2dBasicClass(kref,epsilonRef,mg,s,ec,theta)

{
}

Green2d::Green2d(Complex kref,Complex epsilonRef,double MeshSize, double theta)

  : Green2dBasicClass(kref,epsilonRef,MeshGeometry(MeshSize,MeshSize,1),
                      Symmetry(0,0,0),EComponent(1,1,1),theta)

{
}

Tensor Green2d::M()

{ Tensor G=Tensor(zeros);

  Complex I(0,1);
  double REq=MGeom.REq2D();

  Complex F=I*PI/(2*Kpar) * (REq*Hankel1(1,Kpar*REq) + 2*I/(PI*Kpar));

  G.x.x=G.y.y= F * (1-pow(Kpar/Kref,2)/2);
  G.z.z=       F * (1-pow(Kz  /Kref,2));

  return G;
}

Complex FGreen2d::Hankel1_0(double r)

{ FiltHankel1 F1=FiltHankel1(Kpar,KF);
  return F1.FHankelOrder0(r);
}

Complex FGreen2d::Hankel1_1(double r)

{ FiltHankel1 F1=FiltHankel1(Kpar,KF);
  return F1.FHankelOrder1(r);
}

Complex FGreen2d::Hankel1_2(double r)

{ FiltHankel1 F1=FiltHankel1(Kpar,KF);
  return F1.FHankelOrder2(r);
}

FGreen2d::FGreen2d(Complex kref,Complex epsilonRef,double kf,
           const MeshGeometry& mg, 
           const Symmetry& s, const EComponent& ec,double theta) :

  Green2dBasicClass(kref,epsilonRef,mg,s,ec,theta)

{ KF=kf;
}

FGreen2d::FGreen2d(Complex kref,Complex epsilonRef,double MeshSize, 
                   double theta)

  : Green2dBasicClass(kref,epsilonRef,MeshGeometry(MeshSize,MeshSize,1),
                      Symmetry(0,0,0),EComponent(1,1,1),theta)

{ KF=PI/MeshSize;
}




Tensor FGreen2d::M()

{ return GreenE0(vect3d(0,0,0),vect3d(0,0,0))*MGeom.Volume();
}

/*****************************************************************************/





