#if !defined(_GREEN_)
#define _GREEN_

#include<Complex.h>
#include"LibNic.h"
#include"Symmetry.h"

//Relative accuracy for the numerical integrations used for the filtered Hankel
//functions
#define AccRelIntegrFiltGreen  1E-9
#define NullDistance 1E-40

/*****************************************************************************/

class MeshGeometry : public Triplet<double>{
 protected:
 public:
  MeshGeometry(){};
  MeshGeometry(double dx, double dy, double dz);
  MeshGeometry(const MeshGeometry& mg);
  double Volume() const;
  double REq2D()  const {return sqrt(Volume()/PI);};
  double REq3D()  const {return pow(3.0/4.0/PI*Volume(),1.0/3.0);};
}; 

class EComponent{
 public:
  int N;
  int X; int Y; int Z;
  EComponent(){};
  EComponent(int x, int y, int z);
  EComponent(const EComponent& ec);
};
  
/*****************************************************************************/

class GreenTensor{
 protected:
  Complex      Kref;
  Complex      EpsilonRef;
  MeshGeometry MGeom;
  Symmetry     Symm; 
  EComponent   EComp;
 public:
  GreenTensor(Complex kref,Complex EpsilonRef,const MeshGeometry& mg,
              const Symmetry& s, const EComponent& ec);
// The 4 following function have to be overridden without taking in care
// the symmetry

  virtual Tensor GreenE0(vect3d r,vect3d Rsource);   //Without symmetry
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);   //Without symmetry
  virtual Tensor GreenFFE0(vect3d n,vect3d Rsource); //Without symmetry
  virtual Tensor L();                                //Without symmetry
  virtual Tensor M(){return Tensor(eyes)*MScalar();};
  virtual Tensor MFctMGeom(const MeshGeometry& mgeom)
                    {return Tensor(eyes)*MScalarFctMGeom(mgeom);};

  virtual Complex MScalar(){return MScalarFctMGeom(MGeom);};
  virtual Complex MScalarFctMGeom(const MeshGeometry& mgeom);

// The 3 following function have in principle not to be overridden. It uese the
// overridden fuctions GreenE0 GreenM0 GreenFFE0, takeing in case the 
// symmetries.

  virtual Tensor GreenE(vect3d r,vect3d Rsource);    //With symmetry
  virtual Tensor GreenM(vect3d r,vect3d Rsource);    //With symmetry
  virtual Tensor GreenFFE(vect3d n,vect3d Rsource);  //With symmetry
  Symmetry GetSymmetry(){return Symm;};
};

 
// The class GD2D and GD3D uses the function pointed by 
//  DEpsilonFactor       DEpsilon(r)     = DEpsilonFactor(r)     * DEpsilon
//  DEpsilonTildFactor   DEpsilonTild(r) = DEpsilonTildFactor(r) * DEpsilon
extern double (*DEpsilonFactor)    (double R,double A,vect3d r);
extern double (*DEpsilonTildFactor)(double R,double A,vect3d r);



class Green3dBasicClass : public GreenTensor{
 public:
  Green3dBasicClass(Complex kref,Complex EpsilonRef, const MeshGeometry& mg,
                    const Symmetry& s,
                    const EComponent& ec);
  virtual Tensor GreenFFE0(vect3d n,vect3d Rsource);
  virtual Tensor L();
};

class Green3d : public Green3dBasicClass{
 public:
  Green3d(Complex kref,Complex EpsilonRef, const MeshGeometry& mg,
          const Symmetry& s,
          const EComponent& ec);
  Green3d(Complex kref,Complex EpsilonRef, double MeshSize);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Complex MScalarFctMGeom(const MeshGeometry& mgeom);
};

class FGreen3d : public Green3dBasicClass{
  double KF;
 public:
  FGreen3d(Complex kref,Complex EpsilonRef,double kf, const MeshGeometry& mg,
           const Symmetry& s,
           const EComponent& ec);
  FGreen3d(Complex kref,Complex EpsilonRef, double MeshSize);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Complex MScalar();
};

class FGreen3dL : public Green3dBasicClass{
  double KF;
 public:
  FGreen3dL(Complex kref,Complex EpsilonRef,
            double kf, const MeshGeometry& mg, const Symmetry& s,
            const EComponent& ec);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Complex MScalar();
  virtual Tensor L();
};

/*****************************************************************************/

class Green2dBasicClass : public GreenTensor{
 protected:
  double Theta;
  Complex Kpar;
  Complex Kz;

  virtual Complex Hankel1_0(double r);
  virtual Complex Hankel1_1(double r);
  virtual Complex Hankel1_2(double r);
 public:
  Green2dBasicClass(Complex kref,Complex EpsilonRef,const MeshGeometry& mg, 
                    const Symmetry& s, const EComponent& ec,double theta);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Tensor GreenFFE0(vect3d n,vect3d Rsource);
  virtual Tensor L();
};

class Green2d: public Green2dBasicClass{
 protected:
  virtual Complex Hankel1_0(double r);
  virtual Complex Hankel1_1(double r);
  virtual Complex Hankel1_2(double r);
 public:
  Green2d(Complex kref,Complex EpsilonRef,const MeshGeometry& mg, 
          const Symmetry& s, const EComponent& ec,double theta);
  Green2d(Complex kref,Complex EpsilonRef, double MeshSize,double theta);
  virtual Tensor M();
};

class FGreen2d: public Green2dBasicClass{
  double KF;
 protected:
  virtual Complex Hankel1_0(double r);
  virtual Complex Hankel1_1(double r);
  virtual Complex Hankel1_2(double r);
 public:
  FGreen2d(Complex kref,Complex EpsilonRef,double kf,
           const MeshGeometry& mg, 
          const Symmetry& s, const EComponent& ec,double theta);
  FGreen2d(Complex kref,Complex EpsilonRef, double MeshSize,double theta);
  virtual Tensor M();
};

/*****************************************************************************/


class Green2dTMBasicClass : public GreenTensor{
 public:
  Green2dTMBasicClass(Complex kref,Complex EpsilonRef,const MeshGeometry& mg, 
                      const Symmetry& s, const EComponent& ec);
  virtual Tensor GreenFFE0(vect3d n,vect3d Rsource);
  virtual Tensor L();
  virtual Tensor MFctMGeom(const MeshGeometry& mgeom)
                    {Tensor T=Tensor(zeros);
                     T.z.z=MScalarFctMGeom(mgeom);
                     return T;};
  virtual Tensor M(){Tensor T=Tensor(zeros);
                     T.z.z=MScalar();
                     return T;};
};

class Green2dTM : public Green2dTMBasicClass{   
 public:
  Green2dTM(Complex kref,Complex EpsilonRef,
            const MeshGeometry& mg, const Symmetry& s,
            const EComponent& ec);
  Green2dTM(Complex kref,Complex EpsilonRef, double MeshSize);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Complex MScalarFctMGeom(const MeshGeometry& mgeom);
};

class FGreen2dTM : public Green2dTMBasicClass{
  double KF;
 public:
  FGreen2dTM(Complex kref,Complex EpsilonRef,
             double kf , const MeshGeometry& mg, const Symmetry& s,
                const EComponent& ec);
  FGreen2dTM(Complex kref,Complex EpsilonRef, double MeshSize);
  virtual Complex MScalar();
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
};

/*****************************************************************************/

class Green2dTEBasicClass : public GreenTensor{
 public:
  Green2dTEBasicClass(Complex kref,Complex EpsilonRef, const MeshGeometry& mg, const Symmetry& s,
            const EComponent& ec);
  virtual Tensor GreenFFE0(vect3d n,vect3d Rsource);
  virtual Tensor L();
  virtual Tensor MFctMGeom(const MeshGeometry& mgeom)
                    {Tensor T=Tensor(zeros);
                     T.x.x=T.y.y=MScalarFctMGeom(mgeom);
                     return T;};
  virtual Tensor M(){Tensor T=Tensor(zeros);
                     T.x.x=T.y.y=MScalar();
                     return T;};
};

class Green2dTE : public Green2dTEBasicClass{
 public:
  Green2dTE(Complex kref,Complex EpsilonRef, const MeshGeometry& mg, 
            const Symmetry& s,
            const EComponent& ec);
  Green2dTE(Complex kref,Complex EpsilonRef, double MeshSize);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Complex MScalarFctMGeom(const MeshGeometry& mgeom);
};

class FGreen2dTE : public Green2dTEBasicClass{
  double KF;
 public:
  FGreen2dTE(Complex kref,Complex EpsilonRef, double kf, 
             const MeshGeometry& mg, const Symmetry& s,
             const EComponent& ec);
  FGreen2dTE(Complex kref,Complex EpsilonRef, double MeshSize);
  virtual Tensor GreenE0(vect3d r,vect3d Rsource);
  virtual Tensor GreenM0(vect3d r,vect3d Rsource);
  virtual Complex MScalar();
};

/*****************************************************************************/

Complex Hankel1Inf(int order, Complex k, double rn);

class FiltHankel1 : public Object { 
 private:
  Complex K0;
  double KF;
  Complex FZ0(double);
  Complex FZ1(double);
  Complex FZ2(double);
  double R;
  double T0;
 public:
  FiltHankel1(Complex k0,double kF);
  Complex FHankelOrder0(double r);
  Complex FHankelOrder1(double r);
  Complex FHankelOrder2(double r);
};

#endif



